
require 'thread'

class At_email::Threads::Thread_Queue < At_email::Core::Default

  CLASS_DISPLAY_NAME = 'Thread Queue'

  attr_accessor :threads

  def initialize
    super
    if !@properties
      return false
    end
    @threads = {}
    @properties['Display Name'] = CLASS_DISPLAY_NAME
    @properties['Max Threads'] = 2
  end

  def activate
    @properties['Activation Time'] = Time.now
    @properties['State'] = 'Active'
  end

  def queue_thread(code)
    thread_object = At_email::Tasks::TaskThread.new
    thread_id = thread_object['ID']
    @threads[thread_id] = thread_object
    @threads[thread_id].queue(code)
    return thread_id
  end

  def start_thread(thread_id)
    if get_state() === 'Active'
      @threads[thread_id].start
    else
      formatted_property = get_formatted('State')
      log_string = 'Thread cannot be started as this thread queue is not active - ' + formatted_property
      log_event('W', log_string)
    end
  end

  def stop_thread(thread_id)
    @threads[thread_id].stop
  end

end

class At_email::Tasks::TaskThread < At_email::Core::Default

  CLASS_DISPLAY_NAME = 'Thread'

  attr_accessor :thread_ref

  def initialize
    super
    if !@properties
      return false
    end
    @properties['Display Name'] = CLASS_DISPLAY_NAME
  end

  def queue(code)
    log_string = 'Queueing'
    log_event('D', event_data)
    set('Queue Time', Time.now)
    Thread.abort_on_exception = true
    @thread_ref = Thread.new do
      sleep 0.1
      code
    end
  end

  def start
    @properties['Start Time'] = Time.now
  end

  def stop
    @properties['Stop Time'] = Time.now
    Thread.kill(@thread_ref)
  end

end
